@@ -1,10 +1,8 @@
 /*
  *  zzuf - general purpose fuzzer
- *  Copyright (c) 2002, 2007 Sam Hocevar <sam@zoy.org>
+ *  Copyright (c) 2002-2010 Sam Hocevar <sam@hocevar.net>
  *                All Rights Reserved
  *
- *  $Id$
- *
  *  This program is free software. It comes without any warranty, to
  *  the extent permitted by applicable law. You can redistribute it
  *  and/or modify it under the terms of the Do What The Fuck You Want
@@ -18,6 +16,8 @@
 
 #include "config.h"
 
+#define _INCLUDE_POSIX_SOURCE /* for STDERR_FILENO on HP-UX */
+
 #if defined HAVE_STDINT_H
 #   include <stdint.h>
 #elif defined HAVE_INTTYPES_H
@@ -31,36 +31,36 @@
 #include <stdio.h>
 #include <stdlib.h>
 #if defined HAVE_UNISTD_H
-#   include <unistd.h>
+#   include <unistd.h> /* for read(), write(), close() */
 #endif
 #if defined HAVE_REGEX_H
 #   include <regex.h>
 #endif
 #if defined HAVE_WINSOCK2_H
-#   include <winsock2.h>
-#endif
-#if defined HAVE_WINDOWS_H
-#   include <windows.h>
+#   include <winsock2.h> /* for fd_set */
 #endif
 #if defined HAVE_IO_H
 #   include <io.h>
 #endif
 #include <string.h>
-#include <fcntl.h>
 #include <errno.h>
 #include <signal.h>
+#if defined HAVE_SYS_TIME_H
+#   include <sys/time.h>
+#endif
 #if defined HAVE_SYS_WAIT_H
 #   include <sys/wait.h>
 #endif
 #if defined HAVE_SYS_RESOURCE_H
-#   include <sys/resource.h>
+#   include <sys/resource.h> /* for RLIMIT_AS */
 #endif
 
-#include "libzzuf.h"
+#include "common.h"
 #include "opts.h"
 #include "random.h"
 #include "fd.h"
 #include "fuzz.h"
+#include "myfork.h"
 #include "md5.h"
 #include "timer.h"
 
@@ -91,12 +91,7 @@
 #   undef ZZUF_RLIMIT_CPU
 #endif
 
-/* We use file descriptor 17 as the debug channel */
-#define DEBUG_FILENO 17
-#define DEBUG_FILENO_STR "17"
-
 static void loop_stdin(struct opts *);
-static int run_process(struct opts *, int[][2]);
 
 static void spawn_children(struct opts *);
 static void clean_children(struct opts *);
@@ -106,11 +101,7 @@ static void read_children(struct opts *);
 static void setenv(char const *, char const *, int);
 #endif
 #if defined HAVE_WAITPID
-static char const *sig2str(int);
-#endif
-#if defined HAVE_WINDOWS_H
-static int dll_inject(void *, void *);
-static void *get_entry(char const *);
+static char const *sig2name(int);
 #endif
 static void finfo(FILE *, struct opts *, uint32_t);
 #if defined HAVE_REGEX_H
@@ -120,13 +111,6 @@ static char *merge_file(char *, char *);
 static void version(void);
 static void usage(void);
 
-#if defined HAVE_WINDOWS_H
-static inline void addcpy(void *buf, void *x)
-{
-    memcpy(buf, &x, 4);
-}
-#endif
-
 #define ZZUF_FD_SET(fd, p_fdset, maxfd) \
     if(fd >= 0) \
     { \
@@ -146,7 +130,7 @@ int main(int argc, char *argv[])
     char *include = NULL, *exclude = NULL;
     int cmdline = 0;
 #endif
-    int network = 0;
+    int debug = 0, network = 0;
     int i;
 
     _zz_opts_init(opts);
@@ -169,48 +153,50 @@ int main(int argc, char *argv[])
 #   define OPTSTR_RLIMIT_CPU ""
 #endif
 #define OPTSTR "+" OPTSTR_REGEX OPTSTR_RLIMIT_MEM OPTSTR_RLIMIT_CPU \
-                "Ab:B:C:dD:f:F:ij:l:mnp:P:qr:R:s:St:vxhV"
+                "a:Ab:B:C:dD:e:f:F:ij:l:mnp:P:qr:R:s:St:U:vxhV"
 #define MOREINFO "Try `%s --help' for more information.\n"
         int option_index = 0;
         static struct myoption long_options[] =
         {
             /* Long option, needs arg, flag, short option */
-            { "autoinc",     0, NULL, 'A' },
-            { "bytes",       1, NULL, 'b' },
-            { "max-bytes",   1, NULL, 'B' },
+            { "allow",        1, NULL, 'a' },
+            { "autoinc",      0, NULL, 'A' },
+            { "bytes",        1, NULL, 'b' },
+            { "max-bytes",    1, NULL, 'B' },
 #if defined HAVE_REGEX_H
-            { "cmdline",     0, NULL, 'c' },
+            { "cmdline",      0, NULL, 'c' },
 #endif
-            { "max-crashes", 1, NULL, 'C' },
-            { "debug",       0, NULL, 'd' },
-            { "delay",       1, NULL, 'D' },
+            { "max-crashes",  1, NULL, 'C' },
+            { "debug",        0, NULL, 'd' },
+            { "delay",        1, NULL, 'D' },
 #if defined HAVE_REGEX_H
-            { "exclude",     1, NULL, 'E' },
+            { "exclude",      1, NULL, 'E' },
 #endif
-            { "fuzzing",     1, NULL, 'f' },
-            { "stdin",       0, NULL, 'i' },
+            { "fuzzing",      1, NULL, 'f' },
+            { "stdin",        0, NULL, 'i' },
 #if defined HAVE_REGEX_H
-            { "include",     1, NULL, 'I' },
+            { "include",      1, NULL, 'I' },
 #endif
-            { "jobs",        1, NULL, 'j' },
-            { "list",        1, NULL, 'l' },
-            { "md5",         0, NULL, 'm' },
-            { "max-memory",  1, NULL, 'M' },
-            { "network",     0, NULL, 'n' },
-            { "ports",       1, NULL, 'p' },
-            { "protect",     1, NULL, 'P' },
-            { "quiet",       0, NULL, 'q' },
-            { "ratio",       1, NULL, 'r' },
-            { "refuse",      1, NULL, 'R' },
-            { "seed",        1, NULL, 's' },
-            { "signal",      0, NULL, 'S' },
-            { "max-time",    1, NULL, 't' },
-            { "max-cpu",     1, NULL, 'T' },
-            { "verbose",     0, NULL, 'v' },
-            { "check-exit",  0, NULL, 'x' },
-            { "help",        0, NULL, 'h' },
-            { "version",     0, NULL, 'V' },
-            { NULL,          0, NULL,  0  }
+            { "jobs",         1, NULL, 'j' },
+            { "list",         1, NULL, 'l' },
+            { "md5",          0, NULL, 'm' },
+            { "max-memory",   1, NULL, 'M' },
+            { "network",      0, NULL, 'n' },
+            { "ports",        1, NULL, 'p' },
+            { "protect",      1, NULL, 'P' },
+            { "quiet",        0, NULL, 'q' },
+            { "ratio",        1, NULL, 'r' },
+            { "refuse",       1, NULL, 'R' },
+            { "seed",         1, NULL, 's' },
+            { "signal",       0, NULL, 'S' },
+            { "max-time",     1, NULL, 't' },
+            { "max-cputime",  1, NULL, 'T' },
+            { "max-usertime", 1, NULL, 'U' },
+            { "verbose",      0, NULL, 'v' },
+            { "check-exit",   0, NULL, 'x' },
+            { "help",         0, NULL, 'h' },
+            { "version",      0, NULL, 'V' },
+            { NULL,           0, NULL,  0  }
         };
         int c = mygetopt(argc, argv, OPTSTR, long_options, &option_index);
 
@@ -219,6 +205,9 @@ int main(int argc, char *argv[])
 
         switch(c)
         {
+        case 'a': /* --allow */
+            opts->allow = myoptarg;
+            break;
         case 'A': /* --autoinc */
             setenv("ZZUF_AUTOINC", "1", 1);
             break;
@@ -226,6 +215,8 @@ int main(int argc, char *argv[])
             opts->bytes = myoptarg;
             break;
         case 'B': /* --max-bytes */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             opts->maxbytes = atoi(myoptarg);
             break;
 #if defined HAVE_REGEX_H
@@ -234,14 +225,18 @@ int main(int argc, char *argv[])
             break;
 #endif
         case 'C': /* --max-crashes */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             opts->maxcrashes = atoi(myoptarg);
             if(opts->maxcrashes <= 0)
                 opts->maxcrashes = 0;
             break;
         case 'd': /* --debug */
-            setenv("ZZUF_DEBUG", DEBUG_FILENO_STR, 1);
+            debug++;
             break;
         case 'D': /* --delay */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             opts->delay = (int64_t)(atof(myoptarg) * 1000000.0);
             break;
 #if defined HAVE_REGEX_H
@@ -278,6 +273,8 @@ int main(int argc, char *argv[])
             break;
 #endif
         case 'j': /* --jobs */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             opts->maxchild = atoi(myoptarg) > 1 ? atoi(myoptarg) : 1;
             break;
         case 'l': /* --list */
@@ -288,7 +285,8 @@ int main(int argc, char *argv[])
             break;
 #if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_MEM
         case 'M': /* --max-memory */
-            setenv("ZZUF_MEMORY", "1", 1);
+            if(myoptarg[0] == '=')
+                myoptarg++;
             opts->maxmem = atoi(myoptarg);
             break;
 #endif
@@ -306,6 +304,8 @@ int main(int argc, char *argv[])
             opts->quiet = 1;
             break;
         case 'r': /* --ratio */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             tmp = strchr(myoptarg, ':');
             opts->minratio = atof(myoptarg);
             opts->maxratio = tmp ? atof(tmp + 1) : opts->minratio;
@@ -314,23 +314,34 @@ int main(int argc, char *argv[])
             opts->refuse = myoptarg;
             break;
         case 's': /* --seed */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             tmp = strchr(myoptarg, ':');
             opts->seed = atol(myoptarg);
-            opts->endseed = tmp ? tmp[1] ? (uint32_t)atoi(tmp + 1)
-                                         : (uint32_t)-1UL
+            opts->endseed = tmp ? tmp[1] ? (uint32_t)atol(tmp + 1)
+                                         : (uint32_t)-1L
                                 : opts->seed + 1;
             break;
         case 'S': /* --signal */
             setenv("ZZUF_SIGNAL", "1", 1);
             break;
         case 't': /* --max-time */
-            opts->maxtime = (int64_t)(atof(myoptarg) * 1000000.0);
+            if(myoptarg[0] == '=')
+                myoptarg++;
+            opts->maxtime = (int64_t)atoi(myoptarg) * 1000000;
             break;
 #if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_CPU
-        case 'T': /* --max-cpu */
+        case 'T': /* --max-cputime */
+            if(myoptarg[0] == '=')
+                myoptarg++;
             opts->maxcpu = (int)(atof(myoptarg) + 0.5);
             break;
 #endif
+        case 'U': /* --max-usertime */
+            if(myoptarg[0] == '=')
+                myoptarg++;
+            opts->maxusertime = (int64_t)(atof(myoptarg) * 1000000.0);
+            break;
         case 'x': /* --check-exit */
             opts->checkexit = 1;
             break;
@@ -362,6 +373,15 @@ int main(int argc, char *argv[])
         return EXIT_FAILURE;
     }
 
+    if (opts->allow && !network)
+    {
+        fprintf(stderr, "%s: allow option (-a) requires network fuzzing (-n)\n",
+                argv[0]);
+        printf(MOREINFO, argv[0]);
+        _zz_opts_fini(opts);
+        return EXIT_FAILURE;
+    }
+
     _zz_setratio(opts->minratio, opts->maxratio);
     _zz_setseed(opts->seed);
 
@@ -412,6 +432,9 @@ int main(int argc, char *argv[])
         setenv("ZZUF_EXCLUDE", exclude, 1);
 #endif
 
+    setenv("ZZUF_DEBUG", debug ? debug > 1 ? "2" : "1" : "0", 1);
+    setenv("ZZUF_DEBUGFD", DEBUG_FILENO_STR, 1);
+
     if(opts->fuzzing)
         setenv("ZZUF_FUZZING", opts->fuzzing, 1);
     if(opts->bytes)
@@ -420,10 +443,22 @@ int main(int argc, char *argv[])
         setenv("ZZUF_LIST", opts->list, 1);
     if(opts->ports)
         setenv("ZZUF_PORTS", opts->ports, 1);
+    if(opts->allow && opts->allow[0] == '!')
+        setenv("ZZUF_DENY", opts->allow, 1);
+    else if(opts->allow)
+        setenv("ZZUF_ALLOW", opts->allow, 1);
     if(opts->protect)
         setenv("ZZUF_PROTECT", opts->protect, 1);
     if(opts->refuse)
         setenv("ZZUF_REFUSE", opts->refuse, 1);
+#if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_MEM
+    if(opts->maxmem >= 0)
+    {
+        char buf[32];
+        snprintf(buf, 32, "%i", opts->maxmem);
+        setenv("ZZUF_MEMORY", buf, 1);
+    }
+#endif
 
     /* Allocate memory for children handling */
     opts->child = malloc(opts->maxchild * sizeof(struct child));
@@ -451,7 +486,21 @@ int main(int argc, char *argv[])
 
         if(opts->maxcrashes && opts->crashes >= opts->maxcrashes
             && opts->nchild == 0)
+        {
+            if(opts->verbose)
+                fprintf(stderr,
+                        "zzuf: maximum crash count reached, exiting\n");
             break;
+        }
+
+        if(opts->maxtime && _zz_time() - opts->starttime >= opts->maxtime
+            && opts->nchild == 0)
+        {
+            if(opts->verbose)
+                fprintf(stderr,
+                        "zzuf: maximum running time reached, exiting\n");
+            break;
+        }
     }
 
     /* Clean up */
@@ -464,6 +513,7 @@ static void loop_stdin(struct opts *opts)
 {
     uint8_t md5sum[16];
     struct md5 *ctx = NULL;
+    int total = 0;
 
     if(opts->md5)
         ctx = _zz_md5_init();
@@ -474,8 +524,6 @@ static void loop_stdin(struct opts *opts)
         _zz_bytes(opts->bytes);
     if(opts->list)
         _zz_list(opts->list);
-    if(opts->ports)
-        _zz_ports(opts->ports);
     if(opts->protect)
         _zz_protect(opts->protect);
     if(opts->refuse)
@@ -487,12 +535,22 @@ static void loop_stdin(struct opts *opts)
     for(;;)
     {
         uint8_t buf[BUFSIZ];
-        int ret, off = 0, nw = 0;
+        int ret, toread = BUFSIZ, off = 0, nw = 0;
 
-        ret = read(0, buf, BUFSIZ);
+        if(opts->maxbytes >= 0)
+        {
+            if(total >= opts->maxbytes)
+                break;
+            if(total + BUFSIZ >= opts->maxbytes)
+                toread = opts->maxbytes - total;
+        }
+
+        ret = read(0, buf, toread);
         if(ret <= 0)
             break;
 
+        total += ret;
+
         _zz_fuzz(0, buf, ret);
         _zz_addpos(0, ret);
 
@@ -584,10 +642,8 @@ static char *merge_regex(char *regex, char *string)
 
 static void spawn_children(struct opts *opts)
 {
-    int pipes[3][2];
     int64_t now = _zz_time();
-    pid_t pid;
-    int i, j;
+    int i;
 
     if(opts->nchild == opts->maxchild)
         return; /* no slot */
@@ -598,6 +654,9 @@ static void spawn_children(struct opts *opts)
     if(opts->maxcrashes && opts->crashes >= opts->maxcrashes)
         return; /* all jobs crashed */
 
+    if(opts->maxtime && now - opts->starttime >= opts->maxtime)
+        return; /* run time exceeded */
+
     if(opts->delay > 0 && opts->lastlaunch + opts->delay > now)
         return; /* too early */
 
@@ -606,34 +665,15 @@ static void spawn_children(struct opts *opts)
         if(opts->child[i].status == STATUS_FREE)
             break;
 
-    /* Prepare communication pipe */
-    for(j = 0; j < 3; j++)
+    if (myfork(&opts->child[i], opts) < 0)
     {
-        int ret;
-#if defined HAVE_PIPE
-        ret = pipe(pipes[j]);
-#elif defined HAVE__PIPE
-        ret = _pipe(pipes[j], 512, _O_BINARY | O_NOINHERIT);
-#endif
-        if(ret < 0)
-        {
-            perror("pipe");
-            return;
-        }
-    }
-
-    pid = run_process(opts, pipes);
-    if(pid < 0)
+        fprintf(stderr, "error launching `%s'\n", opts->newargv[0]);
+        opts->seed++;
         return;
+    }
 
     /* We’re the parent, acknowledge spawn */
     opts->child[i].date = now;
-    opts->child[i].pid = pid;
-    for(j = 0; j < 3; j++)
-    {
-        close(pipes[j][1]);
-        opts->child[i].fd[j] = pipes[j][0];
-    }
     opts->child[i].bytes = 0;
     opts->child[i].seed = opts->seed;
     opts->child[i].ratio = _zz_getratio();
@@ -680,8 +720,8 @@ static void clean_children(struct opts *opts)
         }
 
         if(opts->child[i].status == STATUS_RUNNING
-            && opts->maxtime >= 0
-            && now > opts->child[i].date + opts->maxtime)
+            && opts->maxusertime >= 0
+            && now > opts->child[i].date + opts->maxusertime)
         {
             if(opts->verbose)
             {
@@ -753,9 +793,21 @@ static void clean_children(struct opts *opts)
 
             finfo(stderr, opts, opts->child[i].seed);
             fprintf(stderr, "signal %i%s%s\n",
-                    WTERMSIG(status), sig2str(WTERMSIG(status)), message);
+                    WTERMSIG(status), sig2name(WTERMSIG(status)), message);
             opts->crashes++;
         }
+        else if (opts->verbose)
+        {
+            finfo(stderr, opts, opts->child[i].seed);
+            if (WIFSIGNALED(status))
+                fprintf(stderr, "signal %i%s\n",
+                        WTERMSIG(status), sig2name(WTERMSIG(status)));
+            else
+                fprintf(stderr, "exit %i\n", WEXITSTATUS(status));
+        }
+#else
+        /* waitpid() is not available. Don't kill the process. */
+        continue;
 #endif
 
         for(j = 0; j < 3; j++)
@@ -855,7 +907,7 @@ static void setenv(char const *name, char const *value, int overwrite)
 #endif
 
 #if defined HAVE_WAITPID
-static char const *sig2str(int signum)
+static char const *sig2name(int signum)
 {
     switch(signum)
     {
@@ -891,287 +943,34 @@ static char const *sig2str(int signum)
 }
 #endif
 
-static int run_process(struct opts *opts, int pipes[][2])
-{
-    char buf[64];
-#if defined HAVE_FORK
-    static int const files[] = { DEBUG_FILENO, STDERR_FILENO, STDOUT_FILENO };
-    char *libpath, *tmp;
-    int pid, j, len = strlen(opts->oldargv[0]);
-#   if defined __APPLE__
-#       define FILENAME "libzzuf.dylib"
-#       define EXTRAINFO ""
-#       define PRELOAD "DYLD_INSERT_LIBRARIES"
-    setenv("DYLD_FORCE_FLAT_NAMESPACE", "1", 1);
-#   elif defined __osf__
-#       define FILENAME "libzzuf.so"
-#       define EXTRAINFO ":DEFAULT"
-#       define PRELOAD "_RLD_LIST"
-#   else
-#       define FILENAME "libzzuf.so"
-#       define EXTRAINFO ""
-#       define PRELOAD "LD_PRELOAD"
-#   endif
-#elif HAVE_WINDOWS_H
-    PROCESS_INFORMATION pinfo;
-    STARTUPINFO sinfo;
-    HANDLE pid;
-    void *epaddr;
-#endif
-    int ret;
-
-#if defined HAVE_FORK
-    /* Fork and launch child */
-    pid = fork();
-    if(pid < -1)
-        perror("fork");
-    if(pid != 0)
-        return pid;
-
-    /* We loop in reverse order so that files[0] is done last,
-     * just in case one of the other dup2()ed fds had the value */
-    for(j = 3; j--; )
-    {
-        close(pipes[j][0]);
-        if(pipes[j][1] != files[j])
-        {
-            dup2(pipes[j][1], files[j]);
-            close(pipes[j][1]);
-        }
-    }
-#endif
-
-#if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_MEM
-    if(opts->maxmem >= 0)
-    {
-        struct rlimit rlim;
-        rlim.rlim_cur = opts->maxmem * 1000000;
-        rlim.rlim_max = opts->maxmem * 1000000;
-        setrlimit(ZZUF_RLIMIT_MEM, &rlim);
-    }
-#endif
-
-#if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_CPU
-    if(opts->maxcpu >= 0)
-    {
-        struct rlimit rlim;
-        rlim.rlim_cur = opts->maxcpu;
-        rlim.rlim_max = opts->maxcpu + 5;
-        setrlimit(ZZUF_RLIMIT_CPU, &rlim);
-    }
-#endif
-
-    /* Set environment variables */
-    sprintf(buf, "%i", opts->seed);
-    setenv("ZZUF_SEED", buf, 1);
-    sprintf(buf, "%g", opts->minratio);
-    setenv("ZZUF_MINRATIO", buf, 1);
-    sprintf(buf, "%g", opts->maxratio);
-    setenv("ZZUF_MAXRATIO", buf, 1);
-
-#if defined HAVE_FORK
-    /* Meaningless but makes sure there is space for everything */
-    libpath = malloc(len + strlen(LIBDIR "/.libs/" FILENAME EXTRAINFO) + 1);
-    strcpy(libpath, opts->oldargv[0]);
-
-    /* Replace "/path/binaryname" with "/path/.libs/libzzuf.$(EXT)"
-     *     and "binaryname" with ".libs/libzzuf.$(EXT)"
-     * Write the result in libpath. */
-    tmp = strrchr(libpath, '/');
-    strcpy(tmp ? tmp + 1 : libpath, ".libs/" FILENAME);
-
-    ret = access(libpath, R_OK);
-    if(ret < 0)
-        strcpy(libpath, LIBDIR "/" FILENAME);
-
-    /* OSF1 only */
-    strcat(libpath, EXTRAINFO);
-
-    /* Do not clobber previous LD_PRELOAD values */
-    tmp = getenv(PRELOAD);
-    if(tmp && *tmp)
-    {
-        char *bigbuf = malloc(strlen(tmp) + strlen(libpath) + 2);
-        sprintf(bigbuf, "%s:%s", tmp, libpath);
-        free(libpath);
-        libpath = bigbuf;
-    }
-
-    setenv(PRELOAD, libpath, 1);
-    free(libpath);
-
-    if(execvp(opts->newargv[0], opts->newargv))
-    {
-        perror(opts->newargv[0]);
-        exit(EXIT_FAILURE);
-    }
-
-    exit(EXIT_SUCCESS);
-    /* no return */
-    return 0;
-#elif HAVE_WINDOWS_H
-    pid = GetCurrentProcess();
-
-    /* Get entry point */
-    epaddr = get_entry(opts->newargv[0]);
-    if(!epaddr)
-        return -1;
-
-    memset(&sinfo, 0, sizeof(sinfo));
-    sinfo.cb = sizeof(sinfo);
-    DuplicateHandle(pid, (HANDLE)_get_osfhandle(pipes[0][1]), pid,
-        /* FIXME */ &sinfo.hStdInput, 0, TRUE, DUPLICATE_SAME_ACCESS);
-    DuplicateHandle(pid, (HANDLE)_get_osfhandle(pipes[1][1]), pid,
-                    &sinfo.hStdError, 0, TRUE, DUPLICATE_SAME_ACCESS);
-    DuplicateHandle(pid, (HANDLE)_get_osfhandle(pipes[2][1]), pid,
-                    &sinfo.hStdOutput, 0, TRUE, DUPLICATE_SAME_ACCESS);
-    sinfo.dwFlags = STARTF_USESTDHANDLES;
-    ret = CreateProcess(NULL, opts->newargv[0], NULL, NULL, FALSE,
-                        CREATE_SUSPENDED, NULL, NULL, &sinfo, &pinfo);
-    if(!ret)
-        return -1;
-
-    /* Insert the replacement code */
-    ret = dll_inject(pinfo.hProcess, epaddr);
-    if(ret < 0)
-    {
-        TerminateProcess(pinfo.hProcess, -1);
-        return -1;
-    }
-
-    ret = ResumeThread(pinfo.hThread);
-    if(ret < 0)
-    {
-        TerminateProcess(pinfo.hProcess, -1);
-        return -1;
-    }
-
-    return (long int)pinfo.hProcess;
-#endif
-}
-
-#if defined HAVE_WINDOWS_H
-static int dll_inject(void *process, void *epaddr)
-{
-    uint8_t old_ep[7];
-    uint8_t new_ep[] = "\xb8<01>\xff\xe0";
-    uint8_t loader[] = "libzzuf.dll\0<0000c>\xb8<14>\x50\xb8<1a>\xff\xd0"
-                       "\xb8\0\0\0\0\x50\xb8\x07\x00\x00\x00\x50\xb8<2d>"
-                       "\x50\xb8<33>\x50\xb8<39>\xff\xd0\x50\xb8<41>\xff"
-                       "\xd0\xb8<48>\xff\xe0";
-    void *lib;
-    uint8_t *loaderaddr;
-    DWORD tmp;
-
-    /* Save the old entry-point code */
-    ReadProcessMemory(process, epaddr, old_ep, 7, &tmp);
-    if(tmp != 7)
-        return -1;
-
-    loaderaddr = VirtualAllocEx(process, NULL, 78, MEM_COMMIT,
-                                PAGE_EXECUTE_READWRITE);
-    if(!loaderaddr)
-        return -1;
-
-    addcpy(new_ep + 0x01, loaderaddr + 0x0c + 7);
-    WriteProcessMemory(process, epaddr, new_ep, 7, &tmp);
-    if(tmp != 7)
-        return -1;
-
-    lib = LoadLibrary("kernel32.dll");
-    if(!lib)
-        return -1;
-
-    memcpy(loader + 0x0c, old_ep, 7);
-    addcpy(loader + 0x14, loaderaddr + 0x00); /* offset for dll string */
-    addcpy(loader + 0x1a, GetProcAddress(lib, "LoadLibraryA"));
-    addcpy(loader + 0x2d, loaderaddr + 0x0c);
-    addcpy(loader + 0x33, epaddr);
-    addcpy(loader + 0x39, GetProcAddress(lib, "GetCurrentProcess"));
-    addcpy(loader + 0x41, GetProcAddress(lib, "WriteProcessMemory"));
-    addcpy(loader + 0x48, epaddr);
-    FreeLibrary(lib);
-
-    WriteProcessMemory(process, loaderaddr, loader, 78, &tmp);
-    if(tmp != 78)
-        return -1;
-
-    return 0;
-}
-
-static void *get_entry(char const *name)
-{
-    PIMAGE_DOS_HEADER dos;
-    PIMAGE_NT_HEADERS nt;
-    void *file, *map, *base;
-
-    file = CreateFile(name, GENERIC_READ, FILE_SHARE_READ,
-                      NULL, OPEN_EXISTING, 0, NULL);
-    if(file == INVALID_HANDLE_VALUE)
-        return NULL;
-
-    map = CreateFileMapping(file, NULL, PAGE_READONLY, 0, 0, NULL);
-    if(!map)
-    {
-        CloseHandle(file);
-        return NULL;
-    }
-
-    base = MapViewOfFile(map, FILE_MAP_READ, 0, 0, 0);
-    if(!base)
-    {
-        CloseHandle(map);
-        CloseHandle(file);
-        return NULL;
-    }
-
-    /* Sanity checks */
-    dos = (PIMAGE_DOS_HEADER)base;
-    nt = (PIMAGE_NT_HEADERS)((char *)base + dos->e_lfanew);
-    if(dos->e_magic != IMAGE_DOS_SIGNATURE
-      || nt->Signature != IMAGE_NT_SIGNATURE
-      || nt->FileHeader.Machine != IMAGE_FILE_MACHINE_I386
-      || nt->OptionalHeader.Magic != 0x10b /* IMAGE_NT_OPTIONAL_HDR32_MAGIC */)
-    {
-        UnmapViewOfFile(base);
-        CloseHandle(map);
-        CloseHandle(file);
-        return NULL;
-    }
-
-    return (void *)(uintptr_t)(nt->OptionalHeader.ImageBase +
-                                 nt->OptionalHeader.AddressOfEntryPoint);
-}
-#endif
-
 static void version(void)
 {
     printf("zzuf %s\n", PACKAGE_VERSION);
-    printf("Copyright (C) 2002, 2007-2008 Sam Hocevar <sam@zoy.org>\n");
-    printf("This is free software.  You may redistribute copies of it under the\n");
-    printf("terms of the Do What The Fuck You Want To Public License, Version 2\n");
-    printf("<http://sam.zoy.org/wtfpl/>.\n");
-    printf("There is NO WARRANTY, to the extent permitted by law.\n");
+    printf("Copyright (C) 2002-2010 Sam Hocevar <sam@hocevar.net>\n");
+    printf("This program is free software. It comes without any warranty, to the extent\n");
+    printf("permitted by applicable law. You can redistribute it and/or modify it under\n");
+    printf("the terms of the Do What The Fuck You Want To Public License, Version 2, as\n");
+    printf("published by Sam Hocevar. See <http://sam.zoy.org/wtfpl/> for more details.\n");
     printf("\n");
-    printf("Written by Sam Hocevar. Report bugs to <sam@zoy.org>.\n");
+    printf("Written by Sam Hocevar. Report bugs to <sam@hocevar.net>.\n");
 }
 
 static void usage(void)
 {
 #if defined HAVE_REGEX_H
-    printf("Usage: zzuf [-AcdimnqSvx] [-s seed|-s start:stop] [-r ratio|-r min:max]\n");
+    printf("Usage: zzuf [-aAcdimnqSvx] [-s seed|-s start:stop] [-r ratio|-r min:max]\n");
 #else
-    printf("Usage: zzuf [-AdimnqSvx] [-s seed|-s start:stop] [-r ratio|-r min:max]\n");
+    printf("Usage: zzuf [-aAdimnqSvx] [-s seed|-s start:stop] [-r ratio|-r min:max]\n");
 #endif
-    printf("              [-f fuzzing] [-D delay] [-j jobs] [-C crashes] [-B bytes]\n");
-    printf("              [-t seconds] ");
+    printf("              [-f mode] [-D delay] [-j jobs] [-C crashes] [-B bytes] [-a list]\n");
+    printf("              [-t seconds]");
 #if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_CPU
-    printf(                           "[-T seconds] ");
+    printf(                          " [-T seconds]");
 #endif
 #if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_MEM
-    printf(                                        "[-M mebibytes] ");
+    printf(                                       " [-M mebibytes]");
 #endif
-    printf(                                                       "[-b ranges] [-p ports]\n");
+    printf(                                                      " [-b ranges] [-p ports]\n");
     printf("              [-P protect] [-R refuse] [-l list]");
 #if defined HAVE_REGEX_H
     printf(                                                " [-I include] [-E exclude]");
@@ -1183,6 +982,8 @@ static void usage(void)
     printf("Run PROGRAM with optional arguments ARGS and fuzz its input.\n");
     printf("\n");
     printf("Mandatory arguments to long options are mandatory for short options too.\n");
+    printf("  -a, --allow <list>        only fuzz network input for IPs in <list>\n");
+    printf("          ... !<list>       do not fuzz network input for IPs in <list>\n");
     printf("  -A, --autoinc             increment seed each time a new file is opened\n");
     printf("  -b, --bytes <ranges>      only fuzz bytes at offsets within <ranges>\n");
     printf("  -B, --max-bytes <n>       kill children that output more than <n> bytes\n");
@@ -1190,7 +991,7 @@ static void usage(void)
     printf("  -c, --cmdline             only fuzz files specified in the command line\n");
 #endif
     printf("  -C, --max-crashes <n>     stop after <n> children have crashed (default 1)\n");
-    printf("  -d, --debug               print debug messages\n");
+    printf("  -d, --debug               print debug messages (twice for more verbosity)\n");
     printf("  -D, --delay               delay between forks\n");
 #if defined HAVE_REGEX_H
     printf("  -E, --exclude <regex>     do not fuzz files matching <regex>\n");
@@ -1204,25 +1005,26 @@ static void usage(void)
     printf("  -l, --list <list>         only fuzz Nth descriptor with N in <list>\n");
     printf("  -m, --md5                 compute the output's MD5 hash\n");
 #if defined HAVE_SETRLIMIT && defined ZZUF_RLIMIT_MEM
-    printf("  -M, --max-memory <n>      maximum child virtual memory size in MB\n");
+    printf("  -M, --max-memory <n>      maximum child virtual memory in MiB (default %u)\n", DEFAULT_MEM);
 #endif
     printf("  -n, --network             fuzz network input\n");
     printf("  -p, --ports <list>        only fuzz network destination ports in <list>\n");
     printf("  -P, --protect <list>      protect bytes and characters in <list>\n");
     printf("  -q, --quiet               do not print children's messages\n");
     printf("  -r, --ratio <ratio>       bit fuzzing ratio (default %g)\n", DEFAULT_RATIO);
-    printf("      --ratio <start:stop>  specify a ratio range\n");
+    printf("          ... <start:stop>  specify a ratio range\n");
     printf("  -R, --refuse <list>       refuse bytes and characters in <list>\n");
     printf("  -s, --seed <seed>         random seed (default %i)\n", DEFAULT_SEED);
-    printf("      --seed <start:stop>   specify a seed range\n");
+    printf("         ... <start:stop>   specify a seed range\n");
     printf("  -S, --signal              prevent children from diverting crashing signals\n");
-    printf("  -t, --max-time <n>        kill children that run for more than <n> seconds\n");
-    printf("  -T, --max-cpu <n>         kill children that use more than <n> CPU seconds\n");
+    printf("  -t, --max-time <n>        stop spawning children after <n> seconds\n");
+    printf("  -T, --max-cputime <n>     kill children that use more than <n> CPU seconds\n");
+    printf("  -U, --max-usertime <n>    kill children that run for more than <n> seconds\n");
     printf("  -v, --verbose             print information during the run\n");
     printf("  -x, --check-exit          report processes that exit with a non-zero status\n");
     printf("  -h, --help                display this help and exit\n");
     printf("  -V, --version             output version information and exit\n");
     printf("\n");
-    printf("Written by Sam Hocevar. Report bugs to <sam@zoy.org>.\n");
+    printf("Written by Sam Hocevar. Report bugs to <sam@hocevar.net>.\n");
 }
 
